import 'dart:math';
import 'package:luhenchang_plugin/map/bean/latlng.dart';
import 'bean/tuple.dart';
/*
 * 关于时间工具
 * @作者路很长
 * @路很长的博客地址 https://blog.csdn.net/m0_37667770/article/details/80903890
 * @路很长github  https://github.com/luhenchang/flutter_study
 */
class LonLatAndXY {
  //@将WGS84经纬度转为大地2000坐标。我们是国家电网项目数据很精确的了。
  //@param B 纬度
  //@param L 经度
  //@param degree //
  //@param withBand 默认=false
  //@return
  Tuple gps84ToXY(double B, double L, double degree, {withBand = false}) {
    List<double> xy = new List();
    xy.add(0);
    xy.add(0);
    double a = 6378137; //椭球长半轴
    double b = 6356752.3142451795; //椭球短半轴
    double e = 0.081819190842621; //第一偏心率
    double eC = 0.0820944379496957; //第二偏心率

    double L0 = 0; //中央子午线经度
    int n = 0; //带号
    if (degree == 6) {
      //6度
      n = ((L + degree / 2) / degree).round().toInt();
      L0 = degree * n - degree / 2;
    } else {
      //3度
      n = (L / degree).round().toInt();
      L0 = degree * n;
    }
    //开始计算
    double radB = B * pi / 180; //纬度(弧度)
    double radL = L * pi / 180; //经度(弧度)
    double deltaL = (L - L0) * pi / 180; //经度差(弧度)
    double N = a * a / b / sqrt(1 + eC * eC * cos(radB) * cos(radB));
    double C1 = 1.0 +
        3.0 / 4 * e * e +
        45.0 / 64 * pow(e, 4) +
        175.0 / 256 * pow(e, 6) +
        11025.0 / 16384 * pow(e, 8);
    double C2 = 3.0 / 4 * e * e +
        15.0 / 16 * pow(e, 4) +
        525.0 / 512 * pow(e, 6) +
        2205.0 / 2048 * pow(e, 8);
    double C3 = 15.0 / 64 * pow(e, 4) +
        105.0 / 256 * pow(e, 6) +
        2205.0 / 4096 * pow(e, 8);
    double C4 = 35.0 / 512 * pow(e, 6) + 315.0 / 2048 * pow(e, 8);
    double C5 = 315.0 / 131072 * pow(e, 8);
    double t = tan(radB);
    double eta = eC * cos(radB);
    double X = a *
        (1 - e * e) *
        (C1 * radB -
            C2 * sin(2 * radB) / 2 +
            C3 * sin(4 * radB) / 4 -
            C4 * sin(6 * radB) / 6 +
            C5 * sin(8 * radB));

    xy[0] = X +
        N *
            sin(radB) *
            cos(radB) *
            pow(deltaL, 2) *
            (1 +
                pow(deltaL * cos(radB), 2) *
                    (5 - t * t + 9 * eta * eta + 4 * pow(eta, 4)) /
                    12 +
                pow(deltaL * cos(radB), 4) *
                    (61 - 58 * t * t + pow(t, 4)) /
                    360) /
            2;
    xy[1] = N *
            deltaL *
            cos(radB) *
            (1 +
                pow(deltaL * cos(radB), 2) * (1 - t * t + eta * eta) / 6 +
                pow(deltaL * cos(radB), 4) *
                    (5 -
                        18 * t * t +
                        pow(t, 4) -
                        14 * eta * eta -
                        58 * eta * eta * t * t) /
                    120) +
        500000; // +n * 1000000;
    return new Tuple(xy[0], xy[1]);
  }

  //@将大地2000转为WGS84
  //高斯投影反算为大地平面。
  // x，y ，高斯平面坐标点
  //L0 通过经纬度来获取中央带所在带的角度
  //return B纬度 , L经度
  LatLng xyTowgs84(double x, double y, double L0) {
    //中央子午线经度
    //WGS-84   椭球体参数
    double a = 6378137.0; //major semi axis
    double efang = 0.0066943799901413; //square of e
    double e2fang = 0.0067394967422764; //suqre of e2

    y = y - 500000;

    //主曲率计算
    double m0, m2, m4, m6, m8;
    m0 = a * (1 - efang);
    m2 = 3.0 / 2.0 * efang * m0;
    m4 = efang * m2 * 5.0 / 4.0;
    m6 = efang * m4 * 7.0 / 6.0;
    m8 = efang * m6 * 9.0 / 8.0;

    //子午线曲率计算
    double a0, a2, a4, a6, a8;
    a0 = m0 + m2 / 2.0 + m4 * 3.0 / 8.0 + m6 * 5.0 / 16.0 + m8 * 35.0 / 128.0;
    a2 = m2 / 2.0 + m4 / 2.0 + m6 * 15.0 / 32.0 + m8 * 7.0 / 16.0;
    a4 = m4 / 8.0 + m6 * 3.0 / 16.0 + m8 * 7.0 / 32.0;
    a6 = m6 / 32.0 + m8 / 16.0;
    a8 = m8 / 128.0;

    double X = x;
    double FBf = 0;
    double Bf0 = X / a0, Bf1 = 0;

    //计算Bf的值，直到满足条件
    while ((Bf0 - Bf1) >= 0.0001) {
      Bf1 = Bf0;
      FBf = -a2*sin(2*Bf0)/2 +a4*sin(4*Bf0)/4 -a6*sin(6*Bf0)/6 + a8*sin(8*Bf0)/8;
      Bf0 = (X-FBf)/a0;
    }
    double Bf = Bf0;
    //计算公式中参数
    double Wf = sqrt(1 - efang * sin(Bf) * sin(Bf));
    double Nf = a / Wf;
    double Mf = a * (1 - efang) / pow(Wf, 3);
    double nffang = e2fang * cos(Bf) * cos(Bf);
    double tf = tan(Bf);
    double B = Bf - tf * y * y / (2 * Mf * Nf) + tf * (5 + 3 * tf * tf + nffang - 9 * nffang * tf * tf) * pow(y, 4) / (24 * Mf * pow(Nf, 3)) - tf * (61 + 90 * tf * tf + 45 * pow(tf, 4)) * pow(y, 6) / (720 * Mf * pow(Nf, 5));
    double l = y/(Nf*cos(Bf)) - (1+2*tf*tf+nffang)*pow(y, 3)/(6*pow(Nf, 3)*cos(Bf)) + (5+28*tf*tf +24*pow(tf, 4))*pow(y, 5)/(120*pow(Nf, 5)*cos(Bf));
    double L = l + L0;
    //转化成为十进制经纬度格式
    var array_B =rad2dms(B);
    var array_L =rad2dms(L);
    double Bdec = dms2dec(array_B);
    double Ldec = dms2dec(array_L);

    return new LatLng(Bdec, Ldec);
  }

  double p = 180.0 / pi * 3600;

  //通过经纬度来获取中央带所在带的角度
  //@param B 纬度
  //@param L 经度
  //@param N 带[3,6带度]
  double gaussLongToDegreen(double B, double L, int N) {
    //计算该地区的中央子午线经度
    double L00 = (L / 3).round() * 3.toDouble();
    double degreen = L00 / 180 * 3.1415926;
    return degreen;
  }

  //将弧度�?�转化为度分��?
  dynamic rad2dms(double rad) {
    List<int> a = [0,0,0];
    double dms = rad * p;
    a[0] = (dms / 3600.0).floor();
    a[1] = ((dms - a[0] * 3600) / 60.0).floor();
    a[2] = ((dms - a[0] * 3600).toInt() - a[1] * 60);
    return a;
  }

  //将度分秒转化为十进制坐标
  double dms2dec(dynamic dms) {
    double dec = 0.0;
    dec = dms[0] + dms[1] / 60.0 + dms[2] / 3600.0;
    return dec;
  }
}
